/*****************************************************************************
 * net_logger.c : net logging plugin for vlc
 *****************************************************************************
 * Copyright (C) 2005 VideoLAN
 * $Id: net_logger.c 
 *
 * Authors: Juan Carlos Montes Costa <jcmontes_es@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include <stdlib.h>                                      /* malloc(), free() */
#include <string.h>

#include <errno.h>                                                 /* ENOMEM */
#include <stdio.h>

#include <vlc/vlc.h>
#include <vlc/intf.h>
#include <vlc/vlc.h>
#include <vlc/sout.h>

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#ifdef WIN32
#   include <winsock2.h>
#   include <ws2tcpip.h>
#   ifndef IN_MULTICAST
#       define IN_MULTICAST(a) IN_CLASSD(a)
#   endif
#else
#   include <sys/socket.h>
#endif

#include "network.h"

#define DEFAULT_NET_LOGGER_PORT 27777

#define MODE_TEXT 0

#define TEXT_HEADER "-- Net logger module started --\n"
#define TEXT_FOOTER "-- Net logger module stopped --\n"


 /*****************************************************************************
 * intf_sys_t: description and status of net log interface
 *****************************************************************************/
struct intf_sys_t
{
    int i_mode;
    vlc_bool_t  b_udp_out;  /* true if we can open udp */
    int         fd; /* handle udp socket */
    msg_subscription_t *p_sub;
};

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static int  Open    ( vlc_object_t * );                  
static void Close   ( vlc_object_t * );
static void Run     ( intf_thread_t * );

static void FlushQueue        ( msg_subscription_t *, intf_thread_t *, int , int );
static void TextPrint         ( const msg_item_t *, intf_thread_t *, int );

/*****************************************************************************
 * Module descriptor
 *****************************************************************************/
static char *mode_list[] = { "text" };
static char *mode_list_text[] = { N_("Text") };


#define NETLOGGER_ADDR_TEXT N_("NetLogger-Addr")
#define NETLOGGER_ADDR_LONGTEXT N_( \
      "Address to send log messages")

#define NETLOGGER_PORT_TEXT N_("NetLogger-Port")
#define NETLOGGER_PORT_LONGTEXT N_( \
      "Port to send log messages")

#define LOGMODE_TEXT N_("Log format")
#define LOGMODE_LONGTEXT N_("Specify the log format. Only \"text\" at this moment")


vlc_module_begin();
    set_description( _("Net logger interface") );

    add_string( "netlogger-addr", "127.0.0.1" , NULL,
               NETLOGGER_ADDR_TEXT,NETLOGGER_ADDR_LONGTEXT, VLC_TRUE);

    add_integer( "netlogger-port", DEFAULT_NET_LOGGER_PORT , NULL,
                 NETLOGGER_PORT_TEXT, NETLOGGER_ADDR_LONGTEXT, VLC_TRUE);

    add_string( "logmode", "text", NULL, LOGMODE_TEXT, LOGMODE_LONGTEXT,
                VLC_FALSE );

    set_capability( "interface", 0 );
    set_callbacks( Open, Close );

vlc_module_end();

/*****************************************************************************
 * Open: initialize and create stuff
 *****************************************************************************/
static int Open( vlc_object_t *p_this )
{   
    intf_thread_t *p_intf = (intf_thread_t *)p_this;
    char *psz_mode, *psz_addr;

	int i_dst_port;

    CONSOLE_INTRO_MSG;
    msg_Info( p_intf, "Using the net-logger interface module..." );

    /* Allocate instance and initialize some members */
    p_intf->p_sys = (intf_sys_t *)malloc( sizeof( intf_sys_t ) );
    if( p_intf->p_sys == NULL )
    {
        msg_Err( p_intf, "out of memory" );
        return -1;
    }

    psz_mode = config_GetPsz( p_intf, "logmode" );
    if( psz_mode )
    {
        if( !strcmp( psz_mode, "text" ) )
        {
            p_intf->p_sys->i_mode = MODE_TEXT;
        }
        else
        {
            msg_Err( p_intf, "invalid log mode `%s', using `text'", psz_mode );
            p_intf->p_sys->i_mode = MODE_TEXT;
        }

        free( psz_mode );
    }
    else
    {
        msg_Warn( p_intf, "no log mode specified, using `text'" );
        p_intf->p_sys->i_mode = MODE_TEXT;
    }


	/* addr */
    psz_addr = config_GetPsz( p_intf, "netlogger-addr" );

	/* port */
	i_dst_port = config_GetInt( p_intf, "netlogger-port" );
	if( i_dst_port <= 0 )
    {
        i_dst_port = DEFAULT_NET_LOGGER_PORT;
    }

	p_intf->p_sys->fd = net_OpenUDP( p_intf, "", 0, psz_addr, i_dst_port );

	/* Check if it is open correctly */
    if( p_intf->p_sys->fd < 0 )
    { 
	  msg_Err( p_intf, "failed to open udp socket, send disabled" );
      p_intf->p_sys->b_udp_out = VLC_FALSE;
      free( p_intf->p_sys );
      free( psz_addr );
      return -1;
    }
	else
	{
      msg_Dbg( p_intf, "netlogger sends messages to '%s:%d'", psz_addr, i_dst_port );
	}


	p_intf->p_sys->p_sub = msg_Subscribe( p_intf );

    free( psz_addr );

    switch( p_intf->p_sys->i_mode )
    {
    case MODE_TEXT:
    default:
        net_Write( p_intf, p_intf->p_sys->fd, TEXT_HEADER, strlen( TEXT_HEADER ) );
        break;
    }

    p_intf->pf_run = Run;

    return 0;
}



/*****************************************************************************
 * Close: destroy interface stuff
 *****************************************************************************/
static void Close( vlc_object_t *p_this )
{       
    intf_thread_t *p_intf = (intf_thread_t *)p_this;
        
    /* Flush the queue and unsubscribe from the message queue */
    FlushQueue( p_intf->p_sys->p_sub, p_intf, p_intf->p_sys->fd,
                p_intf->p_sys->i_mode );
    msg_Unsubscribe( p_intf, p_intf->p_sys->p_sub );

    switch( p_intf->p_sys->i_mode )
    {
    case MODE_TEXT:
    default:
        net_Write( p_intf, p_intf->p_sys->fd, TEXT_FOOTER, strlen( TEXT_FOOTER ) );
        break;
    }

    /* Close the udp connection*/
    if( (p_intf->p_sys->b_udp_out) && (p_intf->p_sys->fd > 0)  )
    {
        net_Close( p_intf->p_sys->fd );
    }

    /* Destroy structure */
    free( p_intf->p_sys );
}

/*****************************************************************************
 * Run: rc thread
 *****************************************************************************
 * This part of the interface is in a separate thread so that we can call
 * exec() from within it without annoying the rest of the program.
 *****************************************************************************/
static void Run( intf_thread_t *p_intf )
{
    while( !p_intf->b_die )
    {
        FlushQueue( p_intf->p_sys->p_sub, 
			        p_intf, 
			        p_intf->p_sys->fd,
                    p_intf->p_sys->i_mode );

        msleep( INTF_IDLE_SLEEP );
    }
}

/*****************************************************************************
 * FlushQueue: flush the message queue into the log file
 *****************************************************************************/
static void FlushQueue( msg_subscription_t *p_sub, intf_thread_t *p_intf, int fd, int i_mode )
{
    int i_start, i_stop;

    vlc_mutex_lock( p_sub->p_lock );
    i_stop = *p_sub->pi_stop;
    vlc_mutex_unlock( p_sub->p_lock );

    if( p_sub->i_start != i_stop )
    {
        /* Append all messages to log file */
        for( i_start = p_sub->i_start;
             i_start != i_stop;
             i_start = (i_start+1) % VLC_MSG_QSIZE )
        {
            switch( i_mode )
            {
            case MODE_TEXT:
            default:
                TextPrint( &p_sub->p_msg[i_start], p_intf, fd );
                break;
            }
        }

        vlc_mutex_lock( p_sub->p_lock );
        p_sub->i_start = i_start;
        vlc_mutex_unlock( p_sub->p_lock );
    }
}

static const char *ppsz_type[4] = { ": ", " error: ",
                                    " warning: ", " debug: " };

static void TextPrint( const msg_item_t *p_msg, intf_thread_t *p_intf, int fd )
{
	char sz_buffer[1024];
	sprintf ( sz_buffer, "%s%s%s\n", p_msg->psz_module, ppsz_type[p_msg->i_type], p_msg->psz_msg);
    net_Write( p_intf, fd, sz_buffer, strlen( sz_buffer ) );
}
